(*
 *  COMS W4115
 *  YOLOP
 *
 *  Lisa Li
 *  Jonathan Liu
 *  Sasha McIntosh
 *
 *)

open Ast
  
module StringMap = Map.Make(String)

(* Symbol table: Information about all the names in scope *)
type env = {
    functionDec : string StringMap.t; 	(* Signature for each function *)
    globalVars   :  prim StringMap.t; 		(* global variables and their types *)
    localVars    :  prim StringMap.t; 		(* locals + function params and their types *)
  }

(* Include the necessary headers for the C++ back end *)
let headers = "#include <Magick++.h>\n" ^ "using namespace Magick;\n\n"


(* Get the function signature *)
let fsigToString fdecl = 
    fdecl.fname ^ "_" ^ String.concat "_" (List.map Ast.string_of_vdecl fdecl.formals) (* This needs to map the var type to the name *)


(* Some helper functions *)

let vdecl_of_vinit = function
  Vinit(v, _) -> (v.vtype, v.vname)

(* Get the type and name from a variable definition *)
let getTypeAndName_vdef = function
  VarDecl (v) -> (v.vtype, v.vname)
 |Varinit(vi) -> vdecl_of_vinit vi

(* let fparamString fdecl =
    String.concat "_" (List.map Ast.string_of_vartype (fdecl.formals).vtype) *)

(* take the list of vars/names into a list of pairs (type, name) *)
(* this is for a variable declared *)
let rec enum_vardec = function
      [] -> []
    | hd::tl -> (hd.vtype, hd.vname) :: enum_vardec tl

(* this is for a variable declared above *)
let rec enum_vardef = function
      [] -> []
    | hd::tl -> ( getTypeAndName_vdef hd) :: enum_vardef tl

let rec enum_func = function
      [] -> []
    | hd::tl -> (fsigToString hd, hd.fname) ::  enum_func tl

(* functions *)
(* val enum : int -> 'a list -> (int * 'a) list *)
let rec enum stride n = function
    [] -> []
  | hd::tl -> (n, hd) :: enum stride (n+stride) tl

(* variable declarations *)

(* This maps pairs in a graph *)
(* val string_map_pairs StringMap 'a -> (int * 'a) list -> StringMap 'a *)
let string_map_pairs map pairs =
  List.fold_left (fun m (i, n) -> StringMap.add n i m) map pairs

(* Translate the AST into a valid C++ program *)
let yolop_to_cpp (globals, functions) out = 

  (* Map the global variables *)
  let global_Vars = string_map_pairs StringMap.empty (enum_vardef globals) in
  let functionDecs = string_map_pairs StringMap.empty (enum_func functions) in

  (* Translate the AST representation of the function*)
  let translate env fdecl = 
    let localVars = enum_vardef fdecl.locals 
    and formalVars = enum_vardec fdecl.formals in
    let env = { env with localVars = string_map_pairs StringMap.empty (localVars @ formalVars)} in 
    (* get vars *)
    (* Begin translation... *)

    (* Expressions *)
    let rec expr e = 
    (match e with
      IntLiteral(i)       ->  string_of_int i
      | StringLiteral(i)    ->  i 
      | Id(x)               -> 
      if ((StringMap.mem x env.localVars) || (StringMap.mem x env.globalVars))
            then x   (* Check locals *)
      else raise (Failure ("Undeclared identifier:  " ^ x ))
      | Binop(e1,o,e2)      ->  
      expr e1 ^ " " ^
          (match o with
              Add -> "+" | Sub -> "-" | Mult -> "*" | Div -> "/"
            | Equal   -> "==" | Neq     -> "!=" | Less    -> "<" | Leq     -> "<=" | Greater -> ">" | Geq -> ">="
            | And -> "&&" | Or -> "||") ^ " " ^ expr e2
      | Assign(var, e)      ->  
      if ((StringMap.mem var env.localVars) || (StringMap.mem var env.globalVars)) 
          then var ^ "=" ^ expr e
       else raise (Failure ("Undeclared identifier: " ^ var))
      (* Built-in Functions *)
      (* look for e in local vars and then in globals *)
      | Call("print",[e]) ->  "printf (%s," ^ expr e ^ ")"
      | ImgImg(x,y)         ->  "Image(" ^ expr x ^ "," ^ expr y ^ ", Color(\"white\");"
      | ImgPixel(r,g,b,o)   ->  "Color(" ^ expr r ^ "," ^ expr g ^ "," ^ expr b ^ "," ^ expr o ^ ");"
      | Readin(i,s,e)       ->   i ^ ".read(" ^ expr s ^ "." ^ expr e ^ ");"
      | Writeout(i,s,e)     ->  expr i ^ ".write(" ^ expr s ^ "." ^ expr e ^ ");"
      | Dimension(i,e)      ->  expr i ^ "." ^ (match e with 
                                          "height" -> "height();"
                                        | "width"  -> "width();"
                                        | _         -> raise (Failure("Must be \"width\" or \"height\".")))
      | Get(i,x,y)          ->  expr i ^ ".pixelColor(" ^ expr x ^ "," ^ expr y ^ ");"
      | Set(i,x,y,p)        ->  expr i ^ ".pixelColor(" ^ expr x ^ "," ^ expr y ^ "," ^ expr p ^ ");"
      | Getc(p,c)           ->  expr p ^ "." ^ (match expr c with
                                                 "r"    ->  "redQuantum();"
                                                | "g"    ->  "greenQuantum();"
                                                | "b"    ->  "blueQuantum();"
                                                | "o"    ->  "alphaQuantum();"
                                                | _     ->  raise(Failure("Must be 'r' 'g' 'b' or 'o'")))
      | Setc(p,c,v)         -> expr p ^ "." ^ (match expr c with
                                                 "r"    -> "redQuantum(" ^ expr v ^ ");"
                                                | "g"    -> "greenQuantum(" ^ expr v ^ ");"
                                                | "b"    -> "blueQuantum(" ^ expr v ^ ");"
                                                | "o"    -> "alphaQuantum(" ^ expr v ^ ");"
                                                | _     ->  raise(Failure("Must be 'r' 'g' 'b' or 'o'")))
      | Call(s, e)          ->  if (StringMap.mem s env.functionDec) (* Look for the function *)
                                then ( s ^ "(" ^ (String.concat "," (List.map
                                expr (List.rev e))) ^ ");")  (* found the function, add the explicit parameters *)
                                else raise (Failure ("Undefined function: " ^ s)) (* didn't find the function *)
      | Not(e)              -> "!" ^ "(" ^ expr e ^ ")"
      | Noexpr              -> "")
    in let rec stmt = function
        Block(stmts)        -> String.concat "" (List.map stmt stmts) ^ "\n"  (* Clean up the statments *)
      | Expr(e)             -> expr e ^ ";\n"
      | Return(e)           -> "return " ^ expr e ^ ";\n"
      | If(e,s,Block([]))   -> "if(" ^ expr e ^ "){\n" ^ stmt s ^ "}\n"
      | If(e,s1,s2)         -> "if(" ^ expr e ^ "){\n" ^ stmt s1 ^ "}\nelse\n{\n" ^ stmt s2 ^ "}\n"
      | For(e1,e2,e3,s)     -> "for(" ^ expr e1 ^ "," ^ expr e2 ^ "," ^ expr e3^ "){\n" ^ stmt s ^ "}\n"
      | While(e,s)          -> "while(" ^ expr e ^ "){\n" ^ stmt s ^ "}\n"
      | Break               -> "break;\n"
      | Continue            -> "continue;\n"
    in let prim = function
        Int     ->  "int"
      | Void -> "void"
      | String  ->  "string"
      | Pixel   ->  "Color"
      | Image   ->  "Image"

    in let func_params_type = function
        Void -> "void"
      | String  ->  "string"
      | Int -> "int"
      | Pixel -> "Color"
      | Image -> "Image"
    in   (prim fdecl.freturn ^ " " ^ fdecl.fname
          ^ if ((List.length fdecl.formals) != 0)
            then ("("
          ^ func_params_type (List.hd fdecl.formals).vtype ^ " " ^ (List.hd fdecl.formals).vname ^ " "
          ^ String.concat "" (List.map (fun formal -> ", " ^ func_params_type formal.vtype ^ " " ^ formal.vname) (List.tl fdecl.formals)) ^ ")\n{\n")
            else "()\n{\n"
            )
          ^ String.concat "" (List.map Ast.string_of_vdef (List.rev fdecl.locals)) ^ "\n"
          ^ stmt (Block fdecl.body) ^ "\n" ^ 
          if ((String.compare fdecl.fname "main") == 0)
        then "    return 0;\n}\n"
          else "\n}\n"
      
    in let env = {
        functionDec = functionDecs;
        globalVars = global_Vars;
        localVars = StringMap.empty } in

    (* Look for main *)

    let _ = try
      (StringMap.find "main" functionDecs)
    with Not_found -> raise (Failure ("Could not find \"main\" function"))

   (* Found main, now finish translation! *)
    in headers ^ 
      String.concat "" (List.map Ast.string_of_vdef (List.rev globals)) ^ "\n" ^
    (String.concat "\n" (List.map (translate env) (List.rev functions))) ^ "\n"
